#!/usr/bin/env python

from ds485d_mock import *
from ds485c_mock import *
from notification_queue import *
from process_pkt import *
from ds485.dsuid import *
from util import list2hexstr
from iptables import *
from testsuite_harness import *

from Queue import Queue, Empty
import unittest
import time, datetime
import sys, os
import traceback, signal
import logging
import asyncore, socket
import random
from threading import Thread, Timer

TIMESTMP_FMT = ("%H:%M:%S") # no %f for msecs
FORMAT = '%(asctime)s.%(msecs)03d %(name)-5s %(message)s'
logging.basicConfig(format=FORMAT, datefmt=TIMESTMP_FMT, level=logging.INFO)

class Grack(unittest.TestCase):
    log = logging.getLogger("testsuite")

    def setUp(self):
        log = self.log
        #log.setLevel(logging.DEBUG)
        log.debug("setUp")
        log.debug("PID %d", os.getpid())
        signal.signal(signal.SIGINT, self.signal_handler)

        try:
            ports = [ 5004 ]
            self.ds485_port = 8945

            os.system("pkill -9 ds485p")
            iptables_unblock(log, ports[0])

            fork_ds485p(self.ds485_port, ports,
                        loglevel = 5, valgrind = True)

            self.select_loop = AsyncCore()
            self.select_loop.start()

            # open the socket after forking
            self.daemons = fire_up_daemons(ports)
            if not self.daemons:
                log.info("")
                log.info("ds485p failed to connected to our fake-buses")
                log.info("")
                seppuku()

            log.info("")
            log.info("ds485p connected to our fake-buses")
            log.info("")

        except Exception, e:
            print traceback.format_exc()
            raise e # TODO is tearDown called?


    def signal_handler(self, signal, frame):
        self.log.warning('You pressed Ctrl+C!')
        seppuku()

    def tearDown(self):
        self.log.info("tearDown")
        kill_ds485p()
        for d in self.daemons:
            d.close()
        self.select_loop.stop()

    def stop(self):
        self.log.info("do stop")
        self.run = False

    #@unittest.skip("")
    def test_nw_partition(self):
        log = self.log
        log.info("")
        log.info("Network partition test")
        log.info("")
        log.info("We simulate a network partition or cable unplug by blocking the port")
        log.info("of the bus connection with iptables.")
        log.info("The ds485 is sandwiched by one bus and one client.")
        log.info("Since there is only one bus, ds485p emits 'disconnected' when")
        log.info("disconnected from the bus, and 'connected' when it reconnects.")
        log.info("If there were more busses it would't work that way. But we")
        log.info("have to trigger on dsm joint/leave events")
        log.info("Connect and disconnect events are catched with our client")
        log.info("and we use it as triggers to block/unblock the port.")
        log.info("")

        class states:
            disconnect = "d"
            wait_disconnect = "wd"
            reconnect = "r"
            wait_reconnect = "wr"
            fetch_devices = "fd" # fetch devices after reconnect

        self.run = True
        tstate = states.disconnect
        countdown = 10 # steady state
        client = None

        try:
            while self.run:
                try:
                    (evt, pkt, src) = notification_queue.get(True, 1)
                    reply = None
                except Empty: # timeout

                    countdown -= 1
                    if countdown > 0:
                        continue

                    if not client:
                        log.info("Start Client")
                        client = Client('localhost', self.ds485_port, ClientState())
                        continue
                    else:
                        log.debug("client connected %d", client.state.connected)

                    cstate = client.state

                    if tstate == states.disconnect:
                        if not cstate.connected:
                            client.push(bytearray(createQueryBusState(0)))
                            continue

                        log.info("")
                        log.info("      simulate disconnect")
                        log.info("")
                        iptables_block(log, 5004)

                        tstate = states.wait_disconnect

                    elif tstate == states.wait_disconnect:
                        if cstate.connected:
                            continue
                        log.info("")
                        log.info("      disconnected")
                        log.info("")

                        countdown = 10
                        tstate = states.reconnect

                    elif tstate == states.reconnect:

                        log.info("")
                        log.info("      simulate reconnect")
                        log.info("")
                        iptables_unblock(log, 5004)

                        tstate = states.wait_reconnect

                    elif tstate == states.wait_reconnect:
                        if not cstate.connected:
                            continue
                        countdown = 10
                        tstate = states.fetch_devices

                    elif tstate == states.fetch_devices:
                        self.run = False
                        log.info("")
                        log.info("              SUCCESS")
                        log.info("")


                #log.debug("evt: %s pkt: %s src: %r", evt,
                #          list2hexstr(pkt) if pkt else None, src)

                if isinstance(src, Connection):

                    if (evt == handle_read):
                        reply = process_pkt(pkt, log, src.state)

                elif isinstance(src, Client):

                    if (evt == handle_connect):
                        reply = createQueryBusState(0)
                    elif evt == handle_close:
                        log.info("server closed connection, reconnect")
                        client = None
                    elif (evt == handle_read):
                        # behave like a proper client
                        reply = process_pkt_client(pkt, log, src.state)


                if not reply:
                    continue

                #log.info("reply: %s",  list2hexstr(reply))
                src.push(bytearray(reply))
                reply = None

            if client:
                client.close()

        except KeyboardInterrupt, e:
            print traceback.format_exc()
            pass
        except AssertionError, e:
            raise
        except Exception, e:
            print traceback.format_exc()
            pass


        log.info("exiting...")
        self.assertTrue(True)

if __name__== "__main__":
    unittest.main()
    #seppuku()
